/* ZBOSS Zigbee 3.0
 *
 * Copyright (c) 2012-2018 DSR Corporation, Denver CO, USA.
 * http://www.dsr-zboss.com
 * http://www.dsr-corporation.com
 * All rights reserved.
 *
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software, with or without modification, must only be used with a Nordic
 *    Semiconductor ASA integrated circuit.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
PURPOSE: Globals data definitionfor APS subsystem
*/

#ifndef ZB_APS_GLOBALS_H
#define ZB_APS_GLOBALS_H 1

/*! @cond internals_doc */
/*! \addtogroup ZB_APS */
/*! @{ */

#include "zb_secur.h"

/* I don't know why aliases for addr mode were declared, but setting it to 0 and 1 causes
   need to convert it back to "normal" addr mode. So, let's declare this aliaces equal to
   common addr mode values.
   According to 2.2.4.3.1 APSME-BIND.request addr mode values 0x00 and 0x02 are reserved,
   only 0x01 and 0x03 are used */
#define ZB_APS_BIND_DST_ADDR_GROUP ZB_APS_ADDR_MODE_16_GROUP_ENDP_NOT_PRESENT
#define ZB_APS_BIND_DST_ADDR_LONG  ZB_APS_ADDR_MODE_64_ENDP_PRESENT

/**
   Binding table search result for dst request
 */
typedef ZB_PACKED_PRE struct zb_aps_bind_dst_record_s
{
  zb_uint8_t            dst_addr_mode;   /*!< destination address mode flag, 0
                                          * - group address, otherwise long
                                          * address plus dest endpoint */
  ZB_PACKED_PRE union
  {
    zb_uint16_t group_addr;                /*!< group address */
    zb_aps_bind_long_dst_addr_t long_addr; /*!< @see zb_asp_long_dst_addr_t */
  } u;

} ZB_PACKED_STRUCT zb_aps_bind_dst_record_t;

#define ZB_APS_SET_TRANS_INDEX(dtbli, tansi) \
  ZG->aps.binding.dst_table[(dtbli)].trans_index[(tansi)/8] |= (1 << ((tansi)%8))

#define ZB_APS_RESET_TRANS_INDEX(dtbli, tansi) \
  ZG->aps.binding.dst_table[(dtbli)].trans_index[(tansi)/8] &= ~(1 << ((tansi)%8))

#define ZB_APS_IS_TRANS_INDEX(dtbli, tansi) \
  (ZG->aps.binding.dst_table[(dtbli)].trans_index[(tansi)/8] & (1 << ((tansi)%8)))

/**
   Global binding table
*/
typedef ZB_PACKED_PRE struct zb_aps_binding_table_s
{
  zb_uint8_t              src_n_elements;                               /*!< Count elements in source table */
  zb_uint8_t              dst_n_elements;                               /*!< Count elements in destination table */
  /* FIXME: why align here? */
  zb_uint8_t              align[2];
#ifndef ZB_CONFIGURABLE_MEM
  zb_uint8_t              trans_table[ZB_APS_BIND_TRANS_TABLE_SIZE];    /*!< Buffers for simultaneous sendings */
  zb_aps_bind_src_table_t src_table[ZB_APS_SRC_BINDING_TABLE_SIZE];     /*!< Source table */
  zb_aps_bind_dst_table_t dst_table[ZB_APS_DST_BINDING_TABLE_SIZE];     /*!< Destination table */
#else
  zb_uint8_t              *trans_table;
  zb_aps_bind_src_table_t *src_table;
  zb_aps_bind_dst_table_t *dst_table;
#endif
} ZB_PACKED_STRUCT zb_aps_binding_table_t;

ZB_ASSERT_IF_NOT_ALIGNED_TO_4(zb_aps_binding_table_t)

/**
  Group table entry
 */
typedef ZB_PACKED_PRE struct zb_aps_group_table_ent_s
{
  zb_uint16_t    group_addr;                                    /*!< Group address */
  zb_uint8_t     endpoints[ZB_APS_ENDPOINTS_IN_GROUP_TABLE];    /*!< Endpoints table */
  zb_uint8_t     n_endpoints;                                   /*!< Count elements in endpoints
                                                                 * table */
  zb_uint8_t aligned[1];
} ZB_PACKED_STRUCT zb_aps_group_table_ent_t;


ZB_RING_BUFFER_DECLARE(zb_aps_grp_up_q, zb_uint8_t, ZB_APS_GROUP_UP_Q_SIZE);

/**
   Group addressing data structure
 */
typedef struct zb_aps_group_table_s
{
  zb_aps_group_table_ent_t groups[ZB_APS_GROUP_TABLE_SIZE]; /*!< APS group table */
  zb_ushort_t              n_groups;                        /*!< # of entries in APS group table */
#ifndef ZB_LITE_APS_DONT_TX_PACKET_TO_MYSELF
  zb_aps_grp_up_q_t        local_dup_q; /*!< queue to be used to pass sending
                                          group addresses packets to myself */
#endif
} zb_aps_group_table_t;


/**
   r21 Table 4.33 Trust Center Policy Values
 */
typedef struct zb_tc_policy_s
{
  zb_bitfield_t bdb_mode:1;     /*!< If 1, device is in BDB mode, else in r21 mode  */

  /* TC policy attributes */
  zb_bitfield_t allow_joins:1;
  zb_bitfield_t use_white_list:1;
#ifdef ZB_SECURITY_INSTALLCODES
  zb_bitfield_t allow_installcodes:1;
#endif
#ifndef ZB_LITE_NO_TRUST_CENTER_REQUIRE_KEY_EXCHANGE
  zb_bitfield_t update_trust_center_link_keys_required:1;
#endif
  zb_bitfield_t allow_tc_rejoins:1;   /*<! whether Trust Center Rejoin is allowed */
  zb_bitfield_t allow_tc_link_key_requests:2;
  zb_bitfield_t allow_application_link_key_requests:2;
  zb_bitfield_t allow_remote_policy_change:1;
  zb_bitfield_t network_key_update_method:2;
  zb_uint32_t   network_key_update_period;

  /* BDB attributes */
  zb_uint16_t bdb_commissioning_group_id; /*!< specifies the identifier of the group on which the initiator applies finding & binding.  */
  zb_uint8_t bdb_commissioning_mode;      /*!< @see zb_bdb_commissioning_mode_mask_t See Table 4  Bits of the bdbCommissioningMode attribute:
0
Touchlink: 0 = Do not attempt Touchlink commissioning 1 = Attempt Touchlink commissioning
1
Network steering: 0 = Do not attempt network steering 1 = Attempt network steering
2
Network formation: 0 = Do not attempt to form a network 1 = Attempt to form a network, according to device type2
3
Finding & binding: 0 = Do not attempt finding & binding 1 = Attempt finding & binding
                                           */
  zb_uint8_t  bdb_commissioning_status; /*!< see zb_bdb_error_codes_e  */
  /* bdbJoiningNodeEui64 not really used: BDB spec supposes it ig global
   * variable and only single note can join at a time */
  /* bdbJoiningNodeNewTCLinkKey not really used: BDB spec supposes it ig global
   * variable and only single note can join at a time */
#ifdef ZB_SECURITY_INSTALLCODES
#define bdb_join_uses_install_code_key  allow_installcodes  /*!< bdbJoinUsesInstallCodeKey
                                                             * is equal to TRUE, the Trust
                                                             * Center only permits a node
                                                             * to join its network if
                                                             * a corresponding install
                                                             * code derived preconfigured
                                                             * link key  associated
                                                             * with the node has been
                                                             * preinstalled   */
  #endif
  zb_bitfield_t bdb_node_is_on_a_network:1; /* was ZG->nwk.handle.joined. */
  zb_bitfield_t bdb_node_join_link_key_type:2; /*!< the type of link key (see
                                                * sub-clause 6.3) with which
                                                * the node was able to decrypt
                                                * the network key. @see zb_bdb_node_join_link_key_type_e  */
  zb_uint32_t   bdb_primary_channel_set;
  zb_uint32_t   bdb_secondary_channel_set;
  zb_uint32_t   v_scan_channels;
  zb_uint32_t   bdb_commissioning_time;
  zb_uint8_t    bdb_scan_duration;
  zb_uint8_t    bdb_commissioning_step;
  zb_time_t     bdb_trust_center_node_join_timeout;
#ifndef ZB_LITE_NO_TRUST_CENTER_REQUIRE_KEY_EXCHANGE
#define bdb_trust_center_require_key_exchange update_trust_center_link_keys_required
#endif
  zb_bitfield_t bdb_tclink_key_exchange_attempts:4; /*!< the number of key establishment attempts that have been made to establish a new link key after joining.  */
  zb_bitfield_t bdb_tclink_key_exchange_attempts_max:4; /*!< the maximum number of key establishment attempts that will be made before giving up on the key establishment.  */

#ifdef ZB_DISTRIBUTED_SECURITY_ON
  zb_bitfield_t is_distributed:1;
#endif
  zb_bitfield_t v_do_primary_scan:3; /*!< a bit more than vDoPromaryScan in
                                      * BDB: really scan & join machine
                                      * state. @see enum bdb_join_machine_e */

  /* node-side security policy */
  zb_bitfield_t accept_new_unsolicited_trust_center_link_key:1;
  zb_bitfield_t accept_new_unsolicited_application_link_key:1;
  zb_bitfield_t bdb_waiting_for_tclk_exchange: 1; /*!< Set to 1 when node start Link Key exchange
                                                   * procedure. Reset to zero when TC LK exchange
                                                   * successfully copmletes or failed */
  zb_bitfield_t bdb_ext_channel_scan:1; /*!< Touchlink performs ext scan if 1 */
  zb_bitfield_t ignore_aps_channel_mask:1; /*!< Non standard, but useful: if 1,
                                            * use hard-coded channels set. if 0,
                                            * mask channels sets by
                                            * aps_channel_mask. To be used to
                                            * debug at single channel, or 2
                                            * channels etc */


  zb_uint8_t    bdb_application_signal;  /* Application signal code to be passed into
                                          * zb_zdo_startup_complete */
  zb_uint8_t    bdb_remove_device_param; /*!< Used to store buffer param with Remove Device. */
#ifdef ZB_BDB_TOUCHLINK
/** @cond touchlink */
  zb_uint8_t    tl_first_channel_rpt;
  zb_uint8_t    tl_channel_i;
/** @endcond */ /* touchlink */
#endif  /* ZB_BDB_TOUCHLINK */
  zb_bitfield_t bdb_force_router_rejoin:1;      /* Force rejoin for the router */
  zb_bitfield_t nfn:1;                     /*!< if 1, device is not factory new and bdb init checks
                                            * are ok - no need to reinit commissioning */
  zb_bitfield_t tclk_valid:1;              /*!< if 1, tclk is valid */

#if defined ZB_ED_FUNC && defined ZB_CONTROL4_NETWORK_SUPPORT
  zb_bitfield_t permit_control4_network:1; /*!< if 1, Joining Control4 Network is permitted */
#endif /* defined ZB_ED_FUNC && defined ZB_CONTROL4_NETWORK_SUPPORT */
  zb_bitfield_t aps_unencrypted_transport_key_join:1;         /*for joining the devices requiring APS unencrypted Transport key*/
}
zb_tc_policy_t;

#ifndef ZB_SECURITY_INSTALLCODES
#define ZB_JOIN_USES_INSTALL_CODE_KEY(is_client) 0
#else
zb_bool_t zdo_secur_must_use_installcode(zb_bool_t is_client);
#define ZB_JOIN_USES_INSTALL_CODE_KEY(is_client) zdo_secur_must_use_installcode(is_client)
#endif

/**
   APS Informational Base memory-resident data structure
 */
typedef struct zb_apsib_s
{
  zb_uint8_t  aps_counter;

/**
   Start (field name) of the APSIB section to be saved in NVRAM
 */
#define APSIB_SAVE_START aps_designated_coordinator

  /* table 2.138 - Startup parameters */
  zb_bitfield_t   aps_designated_coordinator : 1; /*!< This boolean flag indicates whether the
                                            device should assume on startup that it must
                                            become a Zigbee coordinator.  */
  zb_bitfield_t   aps_insecure_join : 1; /*!< A boolean flag, which defaults to TRUE and
                                   indicates whether it is OK to use insecure
                                   join on startup.  */

  zb_bitfield_t   aps_use_nvram:1; /*!< if 1, use nvram (load/save)   */
  zb_bitfield_t   aps_nvram_erase_at_start:1; /*!< if 1, erase nvram at start  */
  zb_bitfield_t   always_rejoin:1; /*!< if 1, forbidden Assotiation if Rewjoin fail */
  zb_bitfield_t   is_tc:1;             /*!< true if this is a trusted center */
  zb_bitfield_t   reserve : 2;

  zb_channel_list_t aps_channel_mask_list; /*!< This is the masks list containing allowable
                                                                    * channels on which the device may attempt
                                                                    * to form or join a network at startup time. */
  zb_ext_pan_id_t aps_use_extended_pan_id; /*!< The 64-bit identifier of the network to join
                                             or form.  */
//#ifndef ZB_COORDINATOR_ONLY
  zb_ieee_addr_t  trust_center_address;    /*!< Trust Center IEEE address */
//#endif

#ifndef ZB_NO_NWK_MULTICAST
  zb_uint8_t      aps_nonmember_radius;     /*!< Non-member radius for NWK multicast, a value of 0x07 is treated as infinity see 3.3.1.8.2 sub-clause */

  /* TODO: What value is required for mcf.max_nonmember_radius, where it should be set? */
  zb_uint8_t      aps_max_nonmember_radius; /*!< Maximum non-member radius value for multicast transmission */
#endif                                      /* ZB_NO_NWK_MULTICAST */
  zb_uint32_t outgoing_frame_counter; /*!< OutgoingFrameCounter for APS security
                                       * Aux header when using standard key */

  zb_uint8_t tc_standard_key[ZB_CCM_KEY_SIZE];      /*!< Trust Center Standard Key */
#ifdef ZB_DISTRIBUTED_SECURITY_ON
  zb_uint8_t tc_standard_distributed_key[ZB_CCM_KEY_SIZE];      /*!< Distributed Standard Key */
#endif
  zb_uint8_t coordinator_version; /* Value from the node descriptor */
  zb_aps_device_key_pair_storage_t aps_device_key_pair_storage; /*!< APS Application Key pair table */

#if defined ZB_COORDINATOR_ROLE && defined ZB_SECURITY_INSTALLCODES
#ifndef ZB_CONFIGURABLE_MEM
  zb_aps_installcode_storage_t installcodes_table[ZB_N_APS_KEY_PAIR_ARR_MAX_SIZE];
#else
  zb_aps_installcode_storage_t *installcodes_table;
#endif
  zb_secur_ic_add_t *installcode_to_add;
#endif
#ifdef ZB_SECURITY_INSTALLCODES
  zb_uint8_t installcode[ZB_CCM_KEY_SIZE+2];
  zb_uint8_t installcode_type;
#endif

#ifdef APS_FRAGMENTATION
  zb_uint8_t aps_interframe_delay;
  /* TODO: Should be array of endpoints */
  zb_uint8_t aps_max_window_size;
#endif
  zb_tc_policy_t tcpolicy;
} zb_apsib_t;

#ifdef ZB_DISTRIBUTED_SECURITY_ON
#define IS_DISTRIBUTED_SECURITY() ZB_BDB().is_distributed
#else
#define IS_DISTRIBUTED_SECURITY() 0
#endif

enum zb_aps_retrans_ent_state_e
{
  ZB_APS_RETRANS_ENT_FREE,
  ZB_APS_RETRANS_ENT_SENT_MAC_NOT_CONFIRMED_ALRM_RUNNING,
  ZB_APS_RETRANS_ENT_SENT_MAC_NOT_CONFIRMED_APS_ACKED_ALRM_RUNNING,
  ZB_APS_RETRANS_ENT_SENT_MAC_CONFIRMED_ALRM_RUNNING,
  ZB_APS_RETRANS_ENT_KILL_AT_MAC_CONFIRM,
  ZB_APS_RETRANS_ENT_FRAG_SENT_MAC_CONFIRMED_AFTER_APS_ACKED,
};

/* Moved zb_aps_retrans_ent_t to zboss_api_internal.h */

typedef struct zb_aps_retrans_s
{
#ifndef ZB_CONFIGURABLE_MEM
  zb_aps_retrans_ent_t hash[ZB_N_APS_RETRANS_ENTRIES];      /*!< */
#else
  zb_aps_retrans_ent_t *hash;
#endif
  zb_uint8_t           n_packets;
} zb_aps_retrans_t;


typedef struct zb_aps_tmp_s
{
  zb_uint8_t            neighbor_table_iterator;   /*!< */
} zb_aps_tmp_t;

#define ZB_APS_SET_ZDO_ED_SCAN_FLAG()   ZG->aps.zdo_ed_scan = 1;
#define ZB_APS_CLEAR_ZDO_ED_SCAN_FLAG() ZG->aps.zdo_ed_scan = 0;
#define ZB_APS_GET_ZDO_ED_SCAN_FLAG()   ZG->aps.zdo_ed_scan

#define ZB_APS_SET_CHANNEL_MANAGER_ED_SCAN_FLAG()   ZG->aps.chm_ed_scan = 1;
#define ZB_APS_CLEAR_CHANNEL_MANAGER_ED_SCAN_FLAG() ZG->aps.chm_ed_scan = 0;
#define ZB_APS_GET_CHANNEL_MANAGER_ED_SCAN_FLAG()   ZG->aps.chm_ed_scan

/* zb_aps_dup_tbl_ent_t moved to zboss_api_core.h */

typedef struct zb_aps_duplicate_s
{
#ifndef ZB_CONFIGURABLE_MEM
  zb_aps_dup_tbl_ent_t   dups_table[ZB_APS_DUPS_TABLE_SIZE]; /*!< Duplicates table */
#else
  zb_aps_dup_tbl_ent_t   *dups_table;
#endif
} zb_aps_duplicate_t;

typedef enum zb_aps_in_frag_states_e
{
  APS_IN_FRAG_EMPTY            = 0x00,
  APS_IN_FRAG_RECEIVING        = 0x01,
  APS_IN_FRAG_WINDOW_RECEIVED  = 0x02,
  APS_IN_FRAG_WINDOW_MERGED    = 0x03,
  APS_IN_FRAG_COMPLETE         = 0x04
} zb_aps_in_frag_states_t;

typedef ZB_PACKED_PRE struct zb_aps_in_fragmented_frame_s
{
#define ZB_APS_IN_FRAGMENTED_FRAME_EMPTY 0xFFFF
/* 07/31/2019 EE CR:MINOR Why not addr ref? */
  zb_uint16_t src_addr;   /* 0xFFFF for empty entries */
  zb_uint8_t aps_counter;
  zb_uint8_t total_blocks_num;
  zb_uint8_t current_window;
  zb_bufid_t window_buffers[ZB_APS_MAX_FRAGMENT_NUM_IN_WINDOW];
  zb_bufid_t buffer;
  ZB_PACKED_PRE struct
  {
    zb_bitfield_t assemble_in_progress:1; /* probably, we don't need that */
    zb_bitfield_t ack_timer_scheduled:1;
    zb_bitfield_t state:3;
  } ZB_PACKED_STRUCT flags;
} ZB_PACKED_STRUCT zb_aps_in_fragment_frame_t;

typedef enum zb_aps_out_frag_stages_e
{
  APS_FRAG_INACTIVE              = 0x00,  /**< Fragmentation queue empty */
  APS_FRAG_RECEIVED              = 0x01,  /**< apsde_data received */
  APS_FRAG_GET_MAX_TRANSFER      = 0x02,  /**< send node descriptor and receive max transfer size */
  APS_FRAG_CHAIN_PREPARED        = 0x03,  /**< prepare chain */
  APS_FRAG_BUFFERS_PREPARED      = 0x04,  /**< prepare buffers */
  APS_FRAG_TRANSMISSION          = 0x05,  /**< fragmented transmission in progress */
  APS_FRAG_FINALIZING            = 0x06   /**< transmission completed, to free chains */
} zb_aps_out_frag_stages_t;

typedef struct zb_aps_out_fragmented_frame_s
{
  zb_uint8_t aps_counter;
  zb_uint8_t total_blocks_num;
  zb_uint8_t blocks_sent_mask;
  zb_uint8_t blocks_retry_mask;
  zb_uint8_t current_window;
  zb_uint8_t block_ref[8];
  zb_uint8_t retry_count;
  /* ZB_APS_RETRANS_ENT_SENT_MAC_NOT_CONFIRMED_ALRM_RUNNING case: block_num and block_ack of
   * received ack */
  zb_uint8_t wait_block_ack;
  zb_uint8_t wait_block_num;
  zb_uint8_t state;
  /* 2019-03-06 CR:MAJOR What is the reason of this field? It is not used anywhere in
   * the code. */
  zb_uint8_t dst_max_in;
  zb_uint8_t data_param;
  zb_bool_t transmission_is_scheduled;
  zb_uint8_t addr_mode;
} zb_aps_out_fragment_frame_t;

#ifdef APS_FRAGMENTATION
#define ZB_APS_INVALID_MAX_TRANS_SIZE 0
#define ZB_APS_INVALID_MAX_BUFFER_SIZE 0
#define ZB_APS_MAX_TRANS_SIZE_TABLE_SIZE 10
/* FIXME: Is 1000 sec enough? Looks like we do not need to update it too frequently. */
#define ZB_APS_MAX_TRANS_SIZE_CLOCK 100
#define ZB_APS_MAX_TRANS_SIZE_TMO (10 * ZB_TIME_ONE_SECOND)

typedef struct zb_aps_max_trans_size_s
{
  zb_uint8_t addr_ref;
  zb_uint16_t max_in_trans_size;
  zb_uint8_t max_buffer_size;
  zb_uint8_t clock;
} zb_aps_max_trans_size_t;
#endif


typedef struct zb_aps_func_selector_s
{
  zb_callback_t start_formation;
  zb_callback_t authhenticate_child_directly;
#ifdef ZB_BDB_MODE
  zb_callback_t bdb_formation;
#endif
} zb_aps_func_selector_t;

#define APS_OUT_FRAG_QUEUE_SIZE 2
ZB_RING_BUFFER_DECLARE(zb_aps_out_frag_q, zb_bufid_t, APS_OUT_FRAG_QUEUE_SIZE);

/**
   APS subsystem globals
 */
typedef struct zb_aps_globals_s
{
  zb_apsib_t             aib;           /*!< AIB parameters */
  zb_aps_binding_table_t binding;       /*!< Binding table */
  zb_aps_retrans_t       retrans;       /*!< Retrans table */

  zb_aps_group_table_t   group;         /*!< Group table */
  zb_aps_tmp_t           tmp;           /*!< */
  zb_aps_duplicate_t     dups;
  zb_aps_func_selector_t selector;

  /* Lets use 1 simultaneous fragment rx and 1 simultaneous fragment tx (does not affect
     non-fragment frames). Specification allows that:

     2.2.8.4.5  Fragmented Transmissions - Reception and Rejection, and Acknowledgements
     If an incoming fragmented transaction is already in progress but the addressing and APS
     counter fields do not match those of the received frame, then the received frame may
     optionally be rejected or handled independently as a further transaction.

     TODO: May extend to 1 fragmented transaction from 1 neighbor (if needed).
 */
#ifdef APS_FRAGMENTATION
#define ZB_APS_MAX_IN_FRAGMENT_TRANSMISSIONS 3
  zb_aps_in_fragment_frame_t in_frag[ZB_APS_MAX_IN_FRAGMENT_TRANSMISSIONS];
  zb_aps_out_fragment_frame_t out_frag;
  zb_aps_max_trans_size_t max_trans_size[ZB_APS_MAX_TRANS_SIZE_TABLE_SIZE];
#endif

  zb_aps_out_frag_q_t    out_frag_q;
  
  zb_bitfield_t          authenticated:1; /*!< if 1, the device is authenticated */
  zb_bitfield_t          dups_alarm_running:1;    /*!< if 1, the dups packet is processed*/
  zb_bitfield_t          max_trans_size_alarm_running:1;    /*!< if 1, we are process dups packet */

#ifdef ZB_APS_USER_PAYLOAD
  zb_aps_user_payload_callback_t aps_user_payload_cb;
#endif
} zb_aps_globals_t;

#ifdef ZB_APS_USER_PAYLOAD
zb_bool_t zb_aps_call_user_payload_cb(zb_uint8_t param);

#define ZB_APS_CALL_USER_PAYLOAD_CB(param) zb_aps_call_user_payload_cb(param)
#else
#define ZB_APS_CALL_USER_PAYLOAD_CB(param) ZB_FALSE
#endif /* ZB_APS_USER_PAYLOAD */

/**
   Get APS counter field from AIB.

   @return APS counter
 */
#define ZB_AIB_APS_COUNTER() ZG->aps.aib.aps_counter

/**
   Increment APS counter AIB field.
*/
#define ZB_AIB_APS_COUNTER_INC() ZG->aps.aib.aps_counter++

#define ZB_AIB() ZG->aps.aib

#define APS_SELECTOR() ZG->aps.selector

#define ZB_AIB_APS_CHANNEL_MASK(idx) ZG->aps.aib.aps_channel_mask_list[(idx)]

/**
   Macro to access TC policy implemented according to r21
 */
#define ZB_TCPOL() ZG->aps.aib.tcpolicy

/**
   Macro to access BDB attributes.

   Partially shared with TC policy, so use same data structure and sometimes
   fields aliases
 */

#define ZB_BDB() ZG->aps.aib.tcpolicy

#ifdef ZB_BDB_MODE

#ifndef ZB_LITE_BDB_ONLY_COMMISSIONING
#define ZB_IN_BDB() ZB_BDB().bdb_mode
#else
#define ZB_IN_BDB() 1
#endif

#ifdef ZB_BDB_TOUCHLINK
#define ZB_IN_BDB_NFN() (!ZB_ZLL_IS_FACTORY_NEW())
#else
#define ZB_IN_BDB_NFN() (ZB_BDB().nfn)
#endif

#else
#define ZB_IN_BDB() 0
#define ZB_IN_BDB_NFN() 0
#endif

#define ZG_APS_BIND_DST_TABLE( _p ) ZG->aps.binding.dst_table[ _p ]

#define ZB_APS_DUPS() ZG->aps.dups

#  define ZB_SE_MODE() 0

/** @def ZB_SE_MODE
 *  @brief if 1, then SE mode is enabled
 */
#  define ZB_SE_MODE() 0

/*! @} */
/*! @endcond */

#endif /* ZB_APS_GLOBALS_H */
