/* ZBOSS Zigbee 3.0
 *
 * Copyright (c) 2012-2018 DSR Corporation, Denver CO, USA.
 * http://www.dsr-zboss.com
 * http://www.dsr-corporation.com
 * All rights reserved.
 *
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software, with or without modification, must only be used with a Nordic
 *    Semiconductor ASA integrated circuit.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
PURPOSE: Security services and routines - internals
*/

#ifndef ZB_SECUR_H
#define ZB_SECUR_H 1

#include "zb_common.h"
#include "zb_mac.h"

/*! @cond internals_doc */
/*! \addtogroup ZB_SECUR */
/*! @{ */


#define ZB_NONCE_LENGTH 13
#define ZB_KEY_LENGTH 16


enum zb_secur_key_attributes_e
{
  ZB_SECUR_PROVISIONAL_KEY = 0,
  ZB_SECUR_UNVERIFIED_KEY = 1,
  ZB_SECUR_VERIFIED_KEY = 2,
  ZB_SECUR_APPLICATION_KEY = 3,  /*!< non-standard - for quick check for !TCLK */
  ZB_SECUR_ANY_KEY_ATTR = 0xff   /*!< Any key attribute - for zb_secur_get_link_key_by_address() */
};

enum zb_secur_aps_link_key_type_e
{
  ZB_SECUR_UNIQUE_KEY = 0,
  ZB_SECUR_GLOBAL_KEY = 1
};


enum zb_bdb_node_join_link_key_type_e
{
  BDB_DEFAULT_GLOBAL_TRUST_CENTER_LINK_KEY = 0,
  BDB_DISTRIBUTED_SECURITY_GLOBAL_LINK_KEY = 1,
  BDB_INSTALL_CODE_DERIVED_PRECONFIGURED_LINK_KEY = 2,
/** @cond touchlink */
  BDB_TOUCHLINK_PRECONFIGURED_LINK_KEY = 3
/** @endcond */ /* touchlink */
};


enum zb_secur_key_sources_e
{
  ZB_SECUR_KEY_SRC_UNKNOWN = 0,     /*!< key source is unknown/doesn't matter */
  ZB_SECUR_KEY_SRC_CBKE = 1         /*!< key was obtained via CBKE (key establishment) */
  /* NOTE: if extended with more key sources, don't forget to adjust
   * - zb_aps_device_key_pair_set_t::key_source
   * - zb_zcl_parsed_hdr_t::key_source
   * - zb_aps_hdr_t::key_source
   */
};


/**
   APS Device Key Pair Set structure

   Structure contain record Application key with attributes (address, global/unque, etc.)
 */
/* Since that structire used to save into nvram, pack it! */
typedef ZB_PACKED_PRE struct zb_aps_secur_common_data_s
{
  zb_uint8_t coordinator_version;
  zb_uint8_t align[3];
} ZB_PACKED_STRUCT zb_aps_secur_common_data_t;

typedef ZB_PACKED_PRE struct zb_aps_device_key_pair_set_s
{
  zb_ieee_addr_t  device_address;               /*!< Partner address */
  zb_uint8_t      link_key[ZB_CCM_KEY_SIZE];         /*!< Link key, see Spec. Not use in current
                                                      * release */
  /* 4.4.1.1[2] Security Processing of Outgoing[Incoming] Frames */
/* Moved to zb_aps_device_key_pair_array_t */
/* This fields were commented out after discussion with EE.
 * ZB spec does not require to save APS key frame counters into NVRAM.
 */
#ifndef ZB_LITE_NO_GLOBAL_VS_UNIQUE_KEYS
  zb_bitfield_t   aps_link_key_type:1; /*!< @ref zb_secur_aps_link_key_type_e  */
#endif
  zb_bitfield_t   key_source:1;        /*!< how the key was obtained @ref zb_secur_key_sources_e */
  zb_bitfield_t   key_attributes:2;    /*!< attributes of the key @ref zb_secur_key_attributes_e */
  zb_bitfield_t   reserved:4;
  zb_uint8_t      align[3];
} ZB_PACKED_STRUCT zb_aps_device_key_pair_set_t;

typedef zb_aps_device_key_pair_set_t zb_aps_device_key_pair_nvram_t;

ZB_ASSERT_IF_NOT_ALIGNED_TO_4(zb_aps_device_key_pair_nvram_t)

#define ZB_APS_DEVICE_KEY_PAIR_CACHED (zb_uint32_t)0x07ffffff

/* zb_aps_device_key_pair_array_t moved to zboss_api_internal.h */

typedef struct zb_aps_device_key_pair_storage_s
{
  zb_uint8_t  nvram_page;
  zb_uint8_t cached_i;
  zb_aps_device_key_pair_set_t cached;
#ifndef ZB_CONFIGURABLE_MEM
  zb_aps_device_key_pair_array_t key_pair_set[ZB_N_APS_KEY_PAIR_ARR_MAX_SIZE]; /*!< APS Application Key pair table */
#else
  zb_aps_device_key_pair_array_t *key_pair_set;
#endif
} zb_aps_device_key_pair_storage_t;

typedef ZB_PACKED_PRE struct zb_aps_device_key_pair_set_nvram_1_0_s
{
  zb_ieee_addr_t  device_address;               /*!< Partner address */
  zb_uint8_t master_key[ZB_CCM_KEY_SIZE];       /*!< Master key */
  zb_uint8_t global_link_key;                   /*!< 1 - Global key, 0 - Unique key  */
} ZB_PACKED_STRUCT zb_aps_device_key_pair_set_nvram_1_0_t;


/**
   APS Installcode structure for NVRAM
 */
typedef ZB_PACKED_PRE struct zb_aps_installcode_nvram_s
{
  zb_ieee_addr_t  device_address;               /*!< Partner address */
  /*AEV:Here data was a filler: align[2]; - remove it for storage of IC type for now.
    options lowest 2 bits [0-1]: 00-48, 01-64, 10-96, 11-128 bits ic type.*/
  zb_uint8_t      options;
  zb_uint8_t      align;
  zb_uint8_t      installcode[ZB_CCM_KEY_SIZE+2];       /*!< 16b installcode +2b crc */
  /*hint:we can remove 2b crc at the end of installcode, but we must keep ic type, thus align to dword and don't change it*/
} ZB_PACKED_STRUCT zb_aps_installcode_nvram_t;

extern const zb_uint8_t zb_ic_size_by_type[];

#define ZB_IC_TYPE_MAX_SIZE 16

#define ZB_IC_SIZE_BY_TYPE(t) ( t < ZB_IC_TYPE_MAX ? zb_ic_size_by_type[t]:(ZB_IC_TYPE_MAX-1) )

#define ZB_IC_GET_TYPE_FROM_REC(r) ( (r)->options & 0x3)
/* zb_aps_installcode_storage_t moved to zboss_api_internal.h */

typedef struct zb_secur_ic_add_s
{
  zb_uint8_t *address;
  zb_uint8_t *ic;
  zb_uint8_t do_update;
  zb_uint8_t type;
} zb_secur_ic_add_t;


/**
   Network key types - see r20 4.4.3.1.1, r21 Table 4.9 StandardKeyType Parameter of the Transport-Key, Verify-Key, and Confirm-Key Primitives
 */
typedef enum zb_secur_key_types_e
{
  ZB_TC_MASTER_KEY_OBSOLETE          = 0,      /*!< reserved */
  ZB_STANDARD_NETWORK_KEY            = 1,      /*!< NWK key */
  ZB_APP_MASTER_KEY_OBSOLETE         = 2,      /*!< reserved */
  ZB_APP_LINK_KEY                    = 3,      /*!< Application link key */
  ZB_TC_LINK_KEY                     = 4,      /*!< Trust-center link key */
  ZB_HIGH_SECUR_NETWORK_KEY_OBSOLETE = 5     /*!< reserved */
} zb_secur_key_types_t;


/**
   RequestKey type values - see r21 Table 4-19 RequestKeyType Values 
 */
typedef enum zb_secur_request_key_types_e
{
  ZB_REQUEST_APP_LINK_KEY            = 2,      /*!< Application link key */
  ZB_REQUEST_TC_LINK_KEY             = 4,      /*!< Trust-center link key */
} zb_secur_request_key_types_t;


/**
   Key id - see r20 4.5.1.1.2, r21 Table 4.31 Encoding of the Key Identifier
   Sub-Field and 4.5.3 Cryptographic Key Hierarchy

 */
typedef enum zb_secur_key_id_e
{
  ZB_SECUR_DATA_KEY          = 0,      /*!< Data key  - Link key without a hash */
  ZB_SECUR_NWK_KEY           = 1,      /*!< NWK key - use NWK key*/
  ZB_SECUR_KEY_TRANSPORT_KEY = 2,      /*!< Transport key - use hash of TC Link key
                                        * with 2. Used to protect NWK key send. */
  ZB_SECUR_KEY_LOAD_KEY      = 3,      /*!< Load key - use hash of TC key with
                                        * 3. Used to protect APS LC send */
  ZB_SECUR_DATA_KEY_N_UNENCR = 0xfe,   /*!< Send 2 copy: encrypted and not
                                         encrypted at APS */
  ZB_NOT_SECUR               = 0xff    /*!< Not use APS security */
} zb_secur_key_id_t;


/**
   Security/rejoin states os the 'status' field of APSME-Update-Device as described in table 4.40
 */
enum zb_secur_upd_device_status_e
{
  ZB_STD_SEQ_SECURED_REJOIN     = 0,    /*!< Device rejoin with standard security */
  ZB_STD_SEQ_UNSECURED_JOIN     = 1,    /*!< Device join without security */
  ZB_DEVICE_LEFT                = 2,    /*!< Device left */
  ZB_STD_SEQ_UNSECURED_REJOIN   = 3,    /*!< Device rejoin without standard security */
  ZB_HIGH_SEQ_SECURED_REJOIN    = 4,    /*!< Device rejoin with high security */
  ZB_HIGH_SEQ_UNSECURED_JOIN    = 5,    /*!< Device join without high security */
  ZB_HIGH_SEQ_UNSECURED_REJOIN  = 7     /*!< Device rejoin without high security */
};

/**
   Auxiliary frame header (4.5.1) for NWK frame and NWK key

   Extended nonce subfield set to 1 (4.3.1.1).
   source_address amd key_seq_number exist.
 */
typedef ZB_PACKED_PRE struct zb_nwk_aux_frame_hdr_s
{
  zb_uint8_t     secur_control;         /*!< Security Control Field */
  zb_uint32_t    frame_counter;         /*!< Frame Counter */
  zb_ieee_addr_t source_address;        /*!< Extended Source */
  zb_uint8_t     key_seq_number;        /*!< Key Sequence Number */
} ZB_PACKED_STRUCT zb_nwk_aux_frame_hdr_t;


/**
   Auxiliary frame header (4.5.1) for APS frame encrypted by NWK key

   Extended nonce subfield set to 0 (4.4.1.1).
   source_address absent, key_seq_number exists.
 */
typedef ZB_PACKED_PRE struct zb_aps_nwk_aux_frame_hdr_s
{
  zb_uint8_t     secur_control;      /*!< Security Control Field */
  zb_uint32_t    frame_counter;      /*!< Frame Counter */
  zb_uint8_t     key_seq_number;     /*!< Key Sequence Number */
} ZB_PACKED_STRUCT zb_aps_nwk_aux_frame_hdr_t;

/**
   Auxiliary frame header (4.5.1) for APS frame encrypted by Data key

   Extended nonce subfield set to 0 (4.4.1.1).
   source_address and key_seq_number are absent.
 */
typedef ZB_PACKED_PRE struct zb_aps_data_aux_frame_hdr_s
{
  zb_uint8_t     secur_control;     /*!< Security Control Field */
  zb_uint32_t    frame_counter;     /*!< Frame Counter */
} ZB_PACKED_STRUCT zb_aps_data_aux_frame_hdr_t;

/**
   Auxiliary frame header (4.5.1) for APS frame encrypted by Data key

   Extended nonce subfield set to 1
   source_address is present, key_seq_number is absent.
 */
typedef ZB_PACKED_PRE struct zb_aps_data_aux_nonce_frame_hdr_s
{
  zb_uint8_t     secur_control;     /*!< Security Control Field */
  zb_uint32_t    frame_counter;     /*!< Frame Counter */
  zb_ieee_addr_t source_address;    /*!< Extended Source */
} ZB_PACKED_STRUCT zb_aps_data_aux_nonce_frame_hdr_t;



/**
   CCM nonce (see 4.5.2.2)
 */
typedef ZB_PACKED_PRE struct zb_secur_ccm_nonce_s
{
  zb_ieee_addr_t source_address;    /*!< Extended Source */
  zb_uint32_t    frame_counter;     /*!< Frame Counter */
  zb_uint8_t     secur_control;     /*!< Security Control Field */
} ZB_PACKED_STRUCT zb_secur_ccm_nonce_t;

/**
 * See 4.5.1.1
 */

#define ZB_NWK_STD_SECUR_CONTROL \
  ( 5 /*security level */ | (ZB_SECUR_NWK_KEY << 3) /* key identifier - NWK key */ | 1 << 5 /* ext nonce */)

#define ZB_APS_NWK_STD_SECUR_CONTROL \
  ( 5 /*security level */ | (ZB_SECUR_NWK_KEY << 3) /* key identifier - NWK key */ | 0 /* ext nonce */)

#define ZB_APS_MASTEP_STD_SECUR_CONTROL \
  ( 5 /*security level */ | (ZB_SECUR_KEY_TRANSPORT_KEY << 3) /* key identifier - APS master key */ | 0 /* ext nonce */)

#define ZB_APS_LINK_STD_SECUR_CONTROL \
  ( 5 /*security level */ | (ZB_SECUR_KEY_LOAD_KEY << 3) /* key identifier - APS link key */ | 0 /* ext nonce */)

#define ZB_APS_DATA_STD_SECUR_CONTROL \
  ( 5 /*security level */ | (ZB_SECUR_DATA_KEY << 3) /* key identifier - APS data key */ | 0 /* ext nonce */)

#define ZB_APS_SET_SECUR_CONTROL(security_level, key_identifier, ext_nonce) \
  ( (security_level) | ((key_identifier) << 3) | (((ext_nonce != 0u) ? 1 : 0) << 5) )

#define ZB_SECUR_SET_ZEROED_LEVEL(ctrl)  (ctrl) = (ctrl) & (~0x07)

#define ZB_SECUR_AUX_HDR_GET_KEY_TYPE(ctrl) (((ctrl) >> 3) & 0x3)

#define ZB_SECUR_AUX_HDR_SET_KEY_TYPE(ctrl, key_type) ( (ctrl) = ((ctrl) & ~(0x3 << 3)) | (((key_type) & 0x3) << 3))

#define ZB_SECUR_GET_SECURITY_LEVEL(ctrl) ((ctrl) & 0x7)

#define ZB_SECUR_SET_SECURITY_LEVEL(ctrl, level) ( (ctrl) = ((ctrl) & ~0x07) | ((level) & 0x07) )

#define ZB_SECUR_GET_SECURITY_NONCE(ctrl) (((ctrl) >> 5) & 0x1)


#if 0
#define ZB_SECUR_SET_BUF_MAC_ENCR(_pbuf) ((_pbuf)->u.hdr.encrypt_type |= ZB_SECUR_MAC_ENCR)

#define ZB_SECUR_SET_BUF_NWK_ENCR(_pbuf) ((_pbuf)->u.hdr.encrypt_type |= ZB_SECUR_NWK_ENCR)

#define ZB_SECUR_SET_BUF_APS_ENCR(_pbuf) ((_pbuf)->u.hdr.encrypt_type |= ZB_SECUR_APS_ENCR)

#define ZB_SECUR_CLR_BUF_MAC_ENCR(_pbuf) ((_pbuf)->u.hdr.encrypt_type &= ~ZB_SECUR_MAC_ENCR)

#define ZB_SECUR_CLR_BUF_NWK_ENCR(_pbuf) ((_pbuf)->u.hdr.encrypt_type &= ~ZB_SECUR_NWK_ENCR)

#define ZB_SECUR_CLR_BUF_APS_ENCR(_pbuf) ((_pbuf)->u.hdr.encrypt_type &= ~ZB_SECUR_APS_ENCR)

#define ZB_SECUR_GET_BUF_MAC_ENCR(_pbuf) (((_pbuf)->u.hdr.encrypt_type & ZB_SECUR_MAC_ENCR) >> 2)

#define ZB_SECUR_GET_BUF_NWK_ENCR(_pbuf) ((_pbuf)->u.hdr.encrypt_type & ZB_SECUR_NWK_ENCR)

#define ZB_SECUR_GET_BUF_APS_ENCR(_pbuf) (((_pbuf)->u.hdr.encrypt_type & ZB_SECUR_APS_ENCR) >> 1)
#endif


/**
   Initialize APS security data structures
 */
void zb_aps_secur_init(void);


void zb_aps_secur_init_after_leave(void);

/**
   Zigbee Keyed Hash Function. Described in Zigbee specification
   section B.1.4, and in FIPS Publication 198. Strictly speaking
   there is nothing about the Keyed Hash Function which restricts
   it to only a single byte input, but that's all Zigbee ever uses.

   The output of this function is an ep_alloced buffer containing
   the key-hashed output, and is garaunteed never to return NULL.

   @param key   - Zigbee Security Key (must be ZBEE_SEC_CONST_KEYSIZE) in length.
   @param input - key index
   @param hash_key - hash key
 */
void zb_cmm_key_hash(zb_uint8_t *key, zb_uint8_t input, zb_uint8_t *hash_key);


/**
 * Size of one hash block.
 * @note THe result of hashing is exactly one hash block.
 */
#define ZBEE_SEC_CONST_BLOCKSIZE 16

/**
 * @brief Zigbee Cryptographic Hash Function, described in Zigbee
 *        specification sections B.1.3 and B.6.
 *
 * This is a Matyas-Meyer-Oseas hash function using the AES-128
 * cipher. The ECB mode of libgcrypt is used to get a raw block
 * cipher.
 *
 * Input may be any length, and the output must be exactly 1-block in length.
 *
 * Implements the function:
 *    Hash(text) = Hash[t];
 *    Hash[0] = 0^(blocksize).
 *    Hash[i] = E(Hash[i-1], M[i]) XOR M[j];
 *    M[i] = i'th block of text, with some padding and flags concatenated.
 *
 * @param input      Hash Input (any length).
 * @param input_len  Hash Input Length.
 * @param output     Hash Output (exactly one block in length).
 */
zb_bool_t zb_sec_b6_hash(zb_uint8_t *input, zb_uint32_t input_len, zb_uint8_t *output);

/**
   Search for link key by pair address and attribute.

   @note Up to 2 keys to the same address can exist: one is
   VERIFIED/PROVISIONING, another is UNVERIFIED.
   See BDB 10.3.3 Behavior when a known node joins.

   @param address - IEEE address
   @param attr - key attribue to search.
   @return Key Pair Set structure
 */
zb_aps_device_key_pair_set_t *zb_secur_get_link_key_by_address(zb_ieee_addr_t address, enum zb_secur_key_attributes_e attr);

/**
   Search for link key index by pair address and attribute.

   @note Up to 2 keys to the same address can exist: one is
   VERIFIED/PROVISIONING, another is UNVERIFIED.
   See BDB 10.3.3 Behavior when a known node joins.

   @param address - IEEE address
   @param attr - key attribue to search.
   @return keypair index if link key exists, -1 otherwise
 */
zb_uint16_t zb_aps_keypair_get_index_by_addr(zb_ieee_addr_t dev_addr, enum zb_secur_key_attributes_e attr);

#ifndef ZB_HW_ZB_AES128
/*
  Encrypt a single 128-bit block by a 128-bit key using AES
  http://en.wikipedia.org/wiki/Advanced_Encryption_Standard
*/
void zb_aes128(zb_uint8_t *key, zb_uint8_t *msg, zb_uint8_t *c);
#endif


#ifndef ZB_HW_ZB_AES128_DEC
void zb_aes128_dec(zb_uint8_t *key, zb_uint8_t *msg, zb_uint8_t *c);
#endif


/**
   CCM* encryption and authentication procedure for Standard security

   This implementation may be not optimal. It done for debug only: be able to
   encrypt in Linux to be able to compare results with real HW.
   To be debugged using test vectors from the spec.


   @param key   - 16-bytes key
   @param nonce - 16-bytes nonce - see 4.5.2.2 "CCM nonce" for details
   @param string_a - 'a' parameter - authentication string -
                     NwkHeader||AuxuluaryHeader for Standard security. See 4.3.1.1 for details
   @param string_a_len - l(a)
   @param string_m - 'm' parameter - text data -
                     Payload for Standard security. See 4.3.1.1 for details
   @param string_m_len - l(m)
   @param crypted_text - encryption result - user supplied buffer. Must have size
          string_m_len + ccm_m rounded to 16.
          Result len is always string_m_len + ccm_m.


   @return RET_OK if success, RET_ERROR in case of error
 */
zb_ret_t
zb_ccm_encrypt_n_auth(
  zb_uint8_t *key,
  zb_uint8_t *nonce,
  zb_uint8_t *string_a,
  zb_ushort_t string_a_len,
  zb_uint8_t *string_m,
  zb_ushort_t string_m_len,
  zb_bufid_t crypted_text);


/**
   CCM* decryption and authentication procedure for Standard security

   @param key   - 16-bytes key
   @param nonce - 16-bytes nonce - see 4.5.2.2 "CCM nonce" for details
   @param buf   - buffer with encrypted_text [IN], for decryption result [OUT]
   @param string_a_len - length string_a - authentication string -
                     NwkHeader||AuxuluaryHeader for Standard security. See 4.3.1.1 for details
   @param string_c_len - length text data -
                     Payload for Standard security. See 4.3.1.1 for details

   @return RET_OK if success, RET_ERROR in case of error (authentication failure)
 */
zb_ret_t
zb_ccm_decrypt_n_auth(
  zb_uint8_t *key,
  zb_uint8_t *nonce,
  zb_bufid_t buf,
  zb_ushort_t string_a_len,
  zb_ushort_t string_c_len);


/**
   Decryption and authentication according to standard security requirements.

   Decrypt "on place". Use HW if possible,
 */
#define zb_ccm_decrypt_n_auth_stdsecur zb_ccm_decrypt_n_auth

void zb_ccm_auth_trans(
  zb_ushort_t ccm_m,
  zb_uint8_t *key,
  zb_uint8_t *nonce,
  zb_uint8_t *string_a,
  zb_uint16_t string_a_len,
  zb_uint8_t *string_m,
  zb_uint16_t string_m_len,
  zb_uint8_t *t);

/*
void encrypt_trans(
  zb_ushort_t ccm_m,
  zb_uint8_t *key,
  zb_uint8_t *nonce,
  zb_uint8_t *string_m,
  zb_uint16_t string_m_len,
  zb_uint8_t *t,
  zb_uint8_t *encrypted);
*/

/**
   Secure frame at NWK level according to the current security mode
   This procedure secures frame 'on place' increasing its size.
   It is supposed that buffer tail has enough free space.

   @param src - nwk packet to secure
   @param mac_hdr_size - size of header placed before mac hdr
   @param dst - buffer for the secured frame


   @return RET_OK if ok, else error code
 */
zb_ret_t zb_nwk_secure_frame(zb_bufid_t src, zb_uint_t mac_hdr_size, zb_bufid_t dst);


/**
   Unsecure NWK frame at NWK level according to the current security mode
   This procedure secures frame 'on place' decreasing its size.

   @param param - ref to nwk packet to unsecure

   @return RET_OK if success or RET_ERROR if unsecure failed
   Attention: if frame unsecure failed, this function reuses packet buffer to
   send NWK status. Don't use the buffer if return code != RET_OK!
 */
zb_ret_t zb_nwk_unsecure_frame(zb_uint8_t param);


/**
   Allocate and fill space for auxiluary security header in the APS data or command frame.

   @param p - pointer to the aux hdr
   @param key_type - key type for aux header see 4.5.1.1.2, @see zb_secur_key_id_t
   @param dst_short_addr - destionation short address
   @param ext_nonce - extended nonce flag

   If nwk_key is false and transport_key is false then use APS data key
 */
void zb_secur_aps_aux_hdr_fill(zb_uint8_t *p, zb_secur_key_id_t key_type, zb_uint16_t dst_short_addr, zb_bool_t ext_nonce);

/**
   Add security to the APS command frame. Allocate space for APS header. Fill APS Aux security header.

   @param buf - buffer with command payload
   @param command_id - command id to assign into APS command header
   @param secure_aps - secure type APS level
   @param dst_short_addr - destination short address
 */
void zb_aps_command_add_secur(zb_bufid_t buf, zb_uint8_t command_id, zb_secur_key_id_t secure_aps, zb_uint16_t dst_short_addr);

/**
   Encrypt APS frame before its send.

   Need to put secured frame into another buffer because original buffer will be
   passed back to APS in CONFIRM primitive and, potentially, will be
   retransmitter by APS layer, so must keep it unencrypted.

   @param src - source buffer with all headers filled, including MHR
   @param mac_hdr_size - size of MAC header
   @param dst - destination buffer (for secure result)
   @param is_tunnel - APS command has tunnel header

   @return RET_OK if successfully encrypted frame, error code otherwise
 */
zb_ret_t zb_aps_secure_frame(zb_bufid_t src, zb_uint_t mac_hdr_size, zb_bufid_t dst, zb_bool_t is_tunnel);


/**
   Decrypt APS frame "on place"

   @param buf - (in/out) buffer holding frame
   @param keypair_i_p - (out) index of APS keypair used to decrypt or (zb_uint16_t)-1 if used default TCLK
   @param key_id_p - pointer to variable that contains key used to protect the frame

   @return RET_OK if successfully decrypted frame, error code otherwise
 */
zb_ret_t zb_aps_unsecure_frame(zb_bufid_t buf, zb_uint16_t *keypair_i_p, zb_secur_key_id_t *key_id_p, zb_bool_t *is_tclk_key);

/**
   Get is_global_key by source address in frame

   @param buf - in buffer holding frame
   @return if has key pair record then is_global_key else 0
 */
zb_uint8_t zb_aps_get_app_link_key_type_from_frame(zb_bufid_t buf);

/**
   Return size of APS aux secure header

   @param secur_control - aux APS header secur_control field
   @return aux header size if ok or -1 if error (unknown key type)
 */
zb_ushort_t zb_aps_secur_aux_size(zb_uint8_t secur_control);


/**
   Encrypt MAC frame before its send.

   Need to put secured frame into another buffer because original buffer will be
   passed back to APS in CONFIRM primitive and, potentially, will be
   retransmitter by APS layer, so must keep it unencrypted.

   @param src - source buffer with all headers filled, including MHR
   @param mac_hdr_size - size of MAC header
   @param dst - destination buffer (for secure result)

   @return RET_OK if successfully encrypted frame, error code otherwise
 */
zb_ret_t zb_mac_secure_frame(zb_bufid_t src, zb_uint_t mac_hdr_size, zb_bufid_t dst);

/**
   Encrypt MAC frame "on place"

   @param param - (in/out) buffer holding frame

   @return RET_OK if successfully decrypted frame, error code otherwise
 */
zb_ret_t zb_mac_unsecure_frame(zb_uint8_t param);

/**
   Parameters for APSME-TRANSPORT-KEY.request primitive
 */
typedef struct zb_apsme_transport_key_req_s
{
  zb_addr_u dest_address;  /*!< destination address  */
  zb_uint8_t addr_mode;    /*!< The type of destination address supplied by
                             the DstAddr parameter - @see zb_addr_mode_e.
                             This field is non-standard: according to table 4.11
                             dest_address can be 64-bit only.
                             But, according to 4.6.3.4.1  Trust Center Operation,
                             TC must issue APSME-TRANSPORT-KEY.request with
                             broadcast address - means, need 16-bit address here!
                           */
  zb_uint8_t key_type;          /*!< @see zb_secur_key_types_t  */
  union key_holder_u {
    struct apsme_transport_key_nwk_s
    {
      zb_uint8_t key[ZB_CCM_KEY_SIZE];     /*!< Key */
      zb_uint8_t key_seq_number;           /*!< Sequence Number */
      zb_uint8_t use_parent;               /*!< Use parent - see 4.6.3 */
      zb_ieee_addr_t parent_address;       /*!< Parent Address */
    } nwk;                                 /*!< Use for transmit NWK key */
    struct apsme_transport_key_app_s
    {
      zb_uint8_t key[ZB_CCM_KEY_SIZE];      /*!< Key */
      zb_ieee_addr_t partner_address;       /*!< Partner address */
      zb_uint8_t initiator;                 /*!< Is initiator.
                                                 @note Exisis in r20, no in r21, seems, still necessary. */
    } app;                                  /*!< Use for transmit Application key */
  } key;
  /* mega-hack: zb_aps_secure_frame uses dest_address to seek APS key. But,
     zb_aps_secure_frame is called from MAC when NWK also put parameter into
     buffer tail. Let's reserve a space for mcps data request parameters.
     Agree, this is ugly solution. The alternative is to parse packet contents,
     but is is more complex: must analyze for tunnel/direct send.
 */
  zb_uint8_t filler[sizeof(zb_mcps_data_req_params_t) > sizeof(union key_holder_u) ? sizeof(zb_mcps_data_req_params_t) - sizeof(union key_holder_u) : 4];
} zb_apsme_transport_key_req_t;

/**
   Variable part of Transport key command for translate NWK key
 */
typedef ZB_PACKED_PRE struct zb_transport_key_nwk_pkt_s
{
  zb_uint8_t     seq_number;        /*!< Sequence Number */
  zb_ieee_addr_t dest_address;      /*!< Destination address */
  zb_ieee_addr_t source_address;    /*!< Source address */
} ZB_PACKED_STRUCT zb_transport_key_nwk_pkt;

/**
   Variable part of Transport key command for translate TC key
 */
typedef ZB_PACKED_PRE struct zb_transport_key_tc_pkt_s
{
  zb_ieee_addr_t dest_address;      /*!< Destination address */
  zb_ieee_addr_t source_address;    /*!< Source address */
} ZB_PACKED_STRUCT zb_transport_key_tc_pkt;

/**
   Variable part of Transport key command for translate Application key
 */
typedef ZB_PACKED_PRE struct zb_transport_key_app_pkt_s
{
  zb_ieee_addr_t partner_address;   /*!< Partner address */
  zb_uint8_t initiator;             /*!< Is initiator */
}  ZB_PACKED_STRUCT zb_transport_key_app_pkt;

/**
   Transport key command data structure


 */
typedef ZB_PACKED_PRE struct zb_transport_key_dsc_pkt_s
{
  zb_uint8_t     key_type;       /*!< indeed, not part of the nwk key descriptor,
                                  * but it simplifies code  */

  zb_uint8_t     key[ZB_CCM_KEY_SIZE];  /*!< Key*/
  ZB_PACKED_PRE union {
    zb_transport_key_nwk_pkt nwk;       /*!< Variable part for translate NWK key */
    zb_transport_key_tc_pkt tc;         /*!< Variable part for translate TC key */
    zb_transport_key_app_pkt app;       /*!< Variable part for translate Application key */
  } key_data;
} ZB_PACKED_STRUCT zb_transport_key_dsc_pkt_t;

/**
   Update device command data structure
 */
typedef ZB_PACKED_PRE struct zb_apsme_update_device_pkt_s
{
  zb_ieee_addr_t device_address;            /*!< 64-bit address of update device */
  zb_uint16_t    device_short_address;      /*!< 16-bit address of update device */
  zb_uint8_t     status;                    /*!< \see zb_secur_upd_device_status_e  */
} ZB_PACKED_STRUCT zb_apsme_update_device_pkt_t;


/**
   APSME-TRANSPORT-KEY.request primitive

   @param param - packet buffer filled be \see zb_transport_key_dsc_pkt_t
 */
void zb_apsme_transport_key_request(zb_uint8_t param);


/**
   Parameter for APSME-UPDATE-DEVICE.request
 */
typedef struct zb_apsme_update_device_req_s
{
  zb_ieee_addr_t dest_address;               /*!< 64-bit Destination address */
  zb_ieee_addr_t device_address;             /*!< 64-bit address of update device */
  zb_uint16_t    device_short_address;       /*!< 16-bit address of update device */
  zb_uint8_t     status;                     /*!< \see zb_secur_upd_device_status_e  */
} zb_apsme_update_device_req_t;



/**
   Parameter for APSME-UPDATE-DEVICE.indication
 */
typedef struct zb_apsme_update_device_ind_s
{
  zb_ieee_addr_t src_address;               /*!< 64-bit Destination address */
  zb_ieee_addr_t device_address;            /*!< 64-bit address of update device */
  zb_uint16_t    device_short_address;      /*!< 16-bit address of update device */
  zb_uint8_t     status;                    /*!< \see zb_secur_upd_device_status_e  */
} zb_apsme_update_device_ind_t;


/**
   Parameter for APSME-SWITCH-KEY.request

   Note about addr_mode: it is spec violation, but it necessary to satisfy TC.
 */
typedef struct zb_apsme_switch_key_req_s
{
  zb_ieee_addr_t dest_address;   /*!< Destination address  */
  zb_uint8_t key_seq_number;     /*!< Sequence Number */
} zb_apsme_switch_key_req_t;


/**
   APSME-SWITCH-KEY.request primitive

   @param param - packet buffer filled be \see zb_apsme_switch_key_req_t
 */
void zb_apsme_switch_key_request(zb_uint8_t param);


/**
   APSME-SWITCH-KEY APS command body
 */
typedef ZB_PACKED_PRE struct zb_apsme_switch_key_pkt_s
{
  zb_uint8_t key_seq_number;     /*!< Key Number */
} ZB_PACKED_STRUCT zb_apsme_switch_key_pkt_t;


/**
   APSME-SWITCH-KEY.indication
 */
typedef struct zb_apsme_switch_key_ind_s
{
  zb_ieee_addr_t src_address;       /*!< Source Address*/
  zb_uint8_t key_seq_number;        /*!< Key Number */
} zb_apsme_switch_key_ind_t;


void zb_aps_in_switch_key(zb_uint8_t param);


/**
   APSME-REMOVE-DEVICE.request primitive parameters structure
 */
typedef ZB_PACKED_PRE struct zb_apsme_remove_device_req_s
{
  zb_ieee_addr_t parent_address;    /*!< Parent Address*/
  zb_ieee_addr_t child_address;     /*!< Child Address*/
} ZB_PACKED_STRUCT zb_apsme_remove_device_req_t;


/**
   Remove device command data structure
 */
typedef ZB_PACKED_PRE struct zb_apsme_remove_device_pkt_s
{
  zb_ieee_addr_t child_address;     /*!< Child Address*/
} ZB_PACKED_STRUCT zb_apsme_remove_device_pkt_t;


/**
   APSME-REMOVE-DEVICE.indication primitive parameters structure
 */
typedef ZB_PACKED_PRE struct zb_apsme_remove_device_ind_s
{
  zb_ieee_addr_t src_address;       /*!< Source Address */
  zb_ieee_addr_t child_address;     /*!< Child Address */
} ZB_PACKED_STRUCT zb_apsme_remove_device_ind_t;

/**
   APSME-REMOVE-DEVICE.request command

   @param param - packet buffer filled be \see zb_apsme_request_key_req_t

   @snippet tp_sec_bv_28_zc.c zb_secur_apsme_remove_device

   See tp_sec_bv-28 sample
 */
void zb_secur_apsme_remove_device(zb_uint8_t param);


/**
   APSME-REMOVE-DEVICE.indication primitive

   @param param - packet buffer filled be \see zb_apsme_remove_device_pkt_t
 */
void zb_aps_in_remove_device(zb_uint8_t param);

/**
   APSME-REQUEST-KEY.request primitive parameters structure
 */
typedef ZB_PACKED_PRE struct zb_apsme_request_key_req_s
{
  zb_ieee_addr_t dest_address;      /*!< Destination address */
  zb_uint8_t     key_type;          /*!< \see zb_secur_request_key_types_e  */
  zb_ieee_addr_t partner_address;   /*!< Partner address */
} ZB_PACKED_STRUCT zb_apsme_request_key_req_t;

typedef ZB_PACKED_PRE struct zb_apsme_request_key_pkt_s
{
  zb_uint8_t     key_type;          /*!< \see zb_secur_request_key_types_e  */
  zb_ieee_addr_t partner_address;   /*!< Partner address */
} ZB_PACKED_STRUCT zb_apsme_request_key_pkt_t;

/**
   APSME-REQUEST-KEY.indication primitive parameters structure
 */
typedef ZB_PACKED_PRE struct zb_apsme_request_key_ind_s
{
  zb_ieee_addr_t src_address;       /*!< Source address */
  zb_uint8_t     key_type;          /*!< \see zb_secur_request_key_types_e  */
  zb_ieee_addr_t partner_address;   /*!< Partner address */

  zb_uint16_t    keypair_i;     /*!< index of keypair used to decrypt cmd */
} ZB_PACKED_STRUCT zb_apsme_request_key_ind_t;

/**
   APSME-VERIFY-KEY.request

   See 4.4.7.1 APSME-VERIFY-KEY.request
 */
typedef struct zb_apsme_verify_key_req_s
{
  zb_ieee_addr_t dest_address;  /*!< Destination address. ONly TC is valid  */
  zb_uint8_t key_type;          /*!< key type. Only ZB_TC_LINK_KEY is legal.  */
} zb_apsme_verify_key_req_t;


/**
4.4.10.7 Verify-Key Command
 */
typedef ZB_PACKED_PRE struct zb_apsme_verify_key_pkt_s
{
  zb_uint8_t key_type;          /*!< key type. Only ZB_TC_LINK_KEY is legal.  */
  zb_ieee_addr_t src_address;  /*!< Source address. */
  zb_uint8_t key_hash[ZB_CCM_KEY_SIZE]; /*!< Initiator Verify-Key Hash Value  */
} ZB_PACKED_STRUCT zb_apsme_verify_key_pkt_t;


/**
4.4.7.2 APSME-VERIFY-KEY.indication
 */
typedef struct zb_apsme_verify_key_ind_s
{
  zb_ieee_addr_t src_address;  /*!< Source address. */
  zb_uint8_t key_type;          /*!< key type. Only ZB_TC_LINK_KEY is legal.  */
  zb_uint8_t key_hash[ZB_CCM_KEY_SIZE]; /*!< Initiator Verify-Key Hash Value  */
} ZB_PACKED_STRUCT zb_apsme_verify_key_ind_t;

/**
4.4.8.1 APSME-CONFIRM-KEY.request
 */
typedef struct zb_apsme_confirm_key_req_s
{
  zb_uint8_t status;          /*!< A value indicating the success or failure of a pre-vious attempt to verify the trust center link key. See Table 2.27*/
  zb_ieee_addr_t dest_address;  /*!< Destination address.   */
  zb_uint8_t key_type;          /*!< key type. Only ZB_TC_LINK_KEY is legal.  */
} zb_apsme_confirm_key_req_t;


/**
4.4.10.8 Confirm-Key Command
 */
typedef ZB_PACKED_PRE struct zb_apsme_confirm_key_pkt_s
{
  zb_uint8_t status;          /*!< A value indicating the success or failure of a pre-vious attempt to verify the trust center link key. See Table 2.27*/
  zb_uint8_t key_type;          /*!< key type. Only ZB_TC_LINK_KEY is legal.  */
  zb_ieee_addr_t dest_address;  /*!< Dest address. */
} ZB_PACKED_STRUCT zb_apsme_confirm_key_pkt_t;


/**
4.4.8.2 APSME-CONFIRM-KEY.indication
 */
typedef struct zb_apsme_confirm_key_ind_s
{
  zb_uint8_t status;          /*!< A value indicating the success or failure of a pre-vious attempt to verify the trust center link key. See Table 2.27*/
  zb_uint8_t key_type;          /*!< key type. Only ZB_TC_LINK_KEY is legal.  */
  zb_ieee_addr_t src_address;  /*!< The extended 64-bit address of the device that sent the verify-key command. */
} zb_apsme_confirm_key_ind_t;



/**
   APSME-REQUEST-KEY.request command

   @param param - packet buffer filled be \see zb_apsme_request_key_req_t
 */
void zb_secur_apsme_request_key(zb_uint8_t param);

/**
   APSME-REQUEST-KEY.indication primitive

   @param param - packet buffer filled be \see zb_apsme_request_key_pkt_t
   @param keypair_i - index of APS key used to decrypt APS packet of -1 if default TCLK
   @param key_id - the key used to protect the frame
 */
void zb_aps_in_request_key(zb_uint8_t param, zb_uint16_t keypair_i
                           , zb_secur_key_id_t key_id
                          );

/**
   APSME-UPDATE-DEVICE.request command

   @param param - packet buffer filled be \see zb_apsme_update_device_req_t
 */
void zb_apsme_update_device_request(zb_uint8_t param);

/**
   Initialize Trust Center functionality
 */
void secur_tc_init(void);


/**
   Generate NWK keys as random numbers
 */
void secur_nwk_generate_keys(void);

/**
   Authenticate child after join
 */
void secur_authenticate_child(zb_uint8_t param);


/**
   Generate NWK key by key sequence number

   @param key_seq_number - key sequence number
   @return ref to NWK key from NWK keys table
 */
zb_uint8_t *secur_nwk_key_by_seq(zb_ushort_t key_seq_number);

/**
  Update AIB Key-Pair table

  @param address - partner address
  @param key - application key
  @param key_type - @see zb_secur_key_types_t
  @param key_attr - @see zb_secur_key_attributes_e
  @param key_source - @see zb_secur_key_sources_e

  @return pointer to the keypair updated/created or NULL in case of error

  @snippet tp_pro_bv-56_zr.c tp_pro_bv-56_zr

  See tp_pro_bv-56 sample

 */
zb_aps_device_key_pair_set_t * zb_secur_update_key_pair(zb_ieee_addr_t address,
                                                        zb_uint8_t* key,
                                                        zb_uint8_t key_type,
                                                        zb_uint8_t key_attr,
                                                        zb_uint8_t key_source
                                                        );

/**
   Get address from link key pair by specified index.

   @param idx - index from link key pair set
   @param dev_addr - pointer to store device address (addr from key pair if found, zero addr otherwise)
 */
void zb_aps_keypair_get_addr_by_idx(zb_uint_t idx, zb_uint8_t *dev_addr);

/**
   Get link key pair by specified index.

   @param idx - index from link key pair set
   @return pointer to the keypair if it is found and NULL otherwise
 */
zb_aps_device_key_pair_set_t* zb_aps_keypair_get_ent_by_idx(zb_uint_t idx);

/**
  Write/add key pair to AIB Key-Pair table

  @param ent - pointer to key pair entry
  @param idx - index from link key pair set to write (or -1 to add new element)

  @return status of operation - RET_OK for success, RET_ERROR otherwise
 */
zb_ret_t zb_aps_keypair_write(zb_aps_device_key_pair_set_t *ent, zb_uint16_t idx);

/**
   Load link key pair by specified index.

   @param idx - index from link key pair set
   @return status of operation - RET_OK for success, RET_NOT_FOUND if key pair is not found
 */
zb_ret_t zb_aps_keypair_load_by_idx(zb_uint_t idx);

/**
   Read link key pair by specified index.

   @param idx - index from link key pair set
   @param ent - pointer to the loaded key pair
   @return status of operation - RET_OK for success, RET_NOT_FOUND if key pair is not found
 */
zb_ret_t zb_aps_keypair_read_by_idx(zb_uint_t idx, zb_aps_device_key_pair_nvram_t *ent);

/**
   Parameters for APSME-TRANSPORT-KEY.indication primitive
 */
typedef struct zb_apsme_transport_key_indication_s
{
  zb_ieee_addr_t src_address;               /*!< Source address */
  zb_uint8_t key_type;                      /*!< @see zb_secur_key_types_t  */
  union {
    struct apsme_transport_key_nwk_ind_s
    {
      zb_uint8_t key[ZB_CCM_KEY_SIZE];      /*!< Key */
      zb_uint8_t key_seq_number;            /*!< Sequence number */
    } nwk;
    struct apsme_transport_key_app_ind_s
    {
      zb_uint8_t key[ZB_CCM_KEY_SIZE];      /*!< app link Key */
      zb_ieee_addr_t partner_address;       /*!< The extended 64-bit address of the device that was also sent this link key. */
    } app;
    struct apsme_transport_key_tc_ind_s
    {
      zb_uint8_t key[ZB_CCM_KEY_SIZE];      /*!< TC LiKey */
    } tc;
  } key;
} zb_apsme_transport_key_indication_t;



/**
   APSME-TRANSPORT-KEY.indication primitive

   @param param - packet buffer filled be \see zb_apsme_transport_key_indication_t
 */
void zb_apsme_transport_key_indication(zb_uint8_t param);


/**
   Switch NWK key

   @param key_number - NWK key number
 */
void secur_nwk_key_switch(zb_uint8_t key_number);

/**
   Switch NWK key and send it via broadcast
 */
void zb_secur_switch_nwk_key_br(zb_uint8_t param);

/**
   Return true if network key is empty (all zeroes).

   @param key - key
   @return ZB_TRUE if key is empty, ZB_FALSE otherwhise
 */
zb_bool_t secur_nwk_key_is_empty(zb_uint8_t *key);


/**
   APSME-REMOVE-DEVICE.indication primitive

   @param param - packet buffer filled be \see zb_apsme_remove_device_ind_t
 */
void zb_apsme_remove_device_indication(zb_uint8_t param);


/**
   APSME-REQUEST-KEY.indication primitive

   @param param - packet buffer filled be \see zb_apsme_request_key_ind_t
 */
void zb_apsme_request_key_indication(zb_uint8_t param);


/**
   APSME-SWITCH-KEY.indication primitive

   @param param - packet buffer filled be \see zb_apsme_switch_key_ind_t
 */
void zb_apsme_switch_key_indication(zb_uint8_t param);


/**
   Send transport key second destination by Request Application key (field 'initiator' = 0)
   Invoke from confirm
 */
zb_bool_t zb_secur_send_transport_key_second(zb_uint8_t param);


/**
   Analize trust frame by command ID, security level and key type

    @param cmd_id - command ID
    @param secured - frame is crypt
    @param is_global_key - key is global

    @return 'frame is trust'
 */
zb_bool_t zb_secur_aps_accept_policy(zb_uint8_t cmd_id, zb_bool_t secured, zb_uint16_t keypair_i);

/**
   APSME-TUNNEL.indication primitive
   Wrap APS command from tunnel command

   @param param - packet buffer filled be and APS command
 */
void zb_aps_in_tunnel_cmd(zb_uint8_t param);

/* Deprecate ? */
zb_bool_t zdo_secur_device_auth_excluded(zb_ieee_addr_t long_addr);

/* Deprecate ? */
void zdo_secur_set_device_auth_excluded(zb_ieee_addr_t long_addr);

/**

 */
void zb_secur_rejoin_after_security_failure(zb_uint8_t param);


#if defined ZB_HW_ZB_AES128 || defined ZB_HW_ZB_AES128 || !defined ZB_SOFT_SECURITY
#include "hw_crypto_api.h"
#endif

void zb_aps_in_verify_key(zb_uint8_t param);

void zb_aps_in_confirm_key(zb_uint8_t param);

void zb_apsme_verify_key_req(zb_uint8_t param);

void zb_zdo_update_tclk(zb_uint8_t param);

void zb_zdo_verify_tclk(zb_uint8_t param);

void zb_apsme_verify_key_indication(zb_uint8_t param);

void zb_apsme_confirm_key_request(zb_uint8_t param);

void zb_apsme_confirm_key_indication(zb_uint8_t param);

void secur_bdb_tc_init(void);

void secur_bdb_init_defaults(void);

void secur_bdb_client_init(void);

void bdb_initiate_key_exchange(zb_uint8_t param);

void bdb_initiate_key_verify(zb_uint8_t param);
zb_bool_t bdb_verify_tclk_in_progress(void);

void bdb_link_key_transport_with_alarm(zb_uint8_t param2, zb_uint16_t param);

zb_uint8_t bdb_cancel_link_key_refresh_alarm(zb_callback_t func, zb_uint8_t param);

void bdb_link_key_refresh_alarm(zb_uint8_t param);

void zdo_authentication_failed(zb_uint8_t param);

/* CR : 04/19/2016 [VP]: Start: put it here to call this functions from apsme_secur.c */
void bdb_request_tclk_alarm(zb_uint8_t param);

void bdb_update_tclk_failed(zb_uint8_t param);
/* CR : 04/19/2016 [VP]: Stop */

void bdb_update_tclk_done(zb_uint8_t param);

void zb_secur_delete_link_keys_by_addr_ref(zb_address_ieee_ref_t addr_ref);

void zb_secur_delete_link_keys_by_long_addr(zb_ieee_addr_t addr);

void zb_secur_delete_link_key_by_idx(zb_uint16_t idx);

zb_ret_t zb_secur_changing_tc_policy_check(void);

zb_aps_device_key_pair_set_t *zb_secur_get_link_key_pair_set(zb_ieee_addr_t address, zb_bool_t valid_only);

zb_aps_device_key_pair_set_t *zb_secur_create_best_suitable_link_key_pair_set(zb_ieee_addr_t address);

zb_ret_t zb_secur_ic_get_key_by_address(zb_ieee_addr_t address, zb_uint8_t *key);

zb_int8_t zb_secur_ic_get_from_tc_storage(zb_ieee_addr_t address, zb_uint8_t *ic_type, zb_uint8_t *buf);

zb_uint8_t *zb_secur_ic_get_from_client_storage(zb_uint8_t* ic_type);

zb_ret_t zb_secur_ic_check_crc(zb_uint8_t ic_type, zb_uint8_t *ic);

zb_ret_t zb_secur_ic_set(zb_uint8_t ic_type, zb_uint8_t *ic);

zb_ret_t zb_secur_ic_from_string(char *ic_str, zb_uint8_t *ic_type, zb_uint8_t *ic);

zb_ret_t zb_secur_ic_str_set(char *ic_str);

zb_ret_t zb_secur_ic_str_add(zb_ieee_addr_t address, char *ic_str);

zb_bool_t zb_secur_aps_cmd_is_encrypted_by_good_key(zb_uint8_t cmd_id, zb_uint16_t src_addr, zb_uint16_t keypair_i);

void zdo_secur_update_tclk_done(zb_uint8_t param);

void secur_generate_key(zb_uint8_t *key);

/*! @} */
/*! @endcond */

zb_bool_t zb_sec_b6_hash_iter_start(void *dev, zb_uint32_t input_addr, zb_uint32_t input_len);
void zb_sec_b6_hash_iter_done(void *dev, zb_uint32_t input_len, zb_uint8_t *calc_hash);

void zdo_initiate_tclk_gen_over_aps(zb_uint8_t param);

void zb_secur_indicate_app_link_key(zb_uint8_t param);

void secur_nwk_generate_key(zb_uint8_t i, zb_uint_t key_seq);

zb_ushort_t zb_secur_gen_upd_dev_status(zb_ushort_t rejoin_network, zb_ushort_t secure_rejoin);

void secur_authenticate_child_directly(zb_uint8_t param);

zb_uint8_t zb_secur_aps_send_policy(zb_uint_t command, zb_uint16_t dest_addr, zb_uint8_t key_type);

zb_bool_t key_for_child_is_found(zb_ieee_addr_t long_addr);

#endif /* ZB_SECUR_H */
